/*-------------------------------------------------------------------------
+-----------------------------------------------------------------------*/

// Sind die Variable DEBUG_FILE definiert, werden die Nachrichten
// mitprotokolliert und zwar in die Datei "DEBUG_FILE"
// Beim Start des Threads wird die Datei neu angelegt!

#define DEBUG_FILE "/var/log/rfsv-server.log"

#ifdef DEBUG_FILE
#define DPRINTF_2(a, b) { printf(a, b); if ((fd_debug = fopen(DEBUG_FILE, "a+")) != NULL) { fprintf(fd_debug, a, b); fclose(fd_debug); } }
#define DPRINTF_3(a, b, c) { printf(a, b, c); if ((fd_debug = fopen(DEBUG_FILE, "a+")) != NULL) { fprintf(fd_debug, a, b, c); fclose(fd_debug); } }
#define DPRINTF_4(a, b, c, d) { printf(a, b, c, d); if ((fd_debug = fopen(DEBUG_FILE, "a+")) != NULL) { fprintf(fd_debug, a, b, c, d); fclose(fd_debug); } }
#endif

#include <sys/stat.h>

#include "Rfsv.h"
#include "mc_rfile.h"
#include "p_file.h"
#include "messages.h"

#define MAX_OPEN_FILES 16
#define OP_BPATH 1
#define OP_BFILE 2
#define OP_BVOL  3
#define OP_BDIR  4
#define MIN_SEND_BUFFER 7

uint16	cid;				// Channelid!

typedef struct {
	bool					aktiv;	/* enthält true, wenn der Eintrag belegt ist!	*/
	int						fd;		/* enthält den Filediskriptor für den PSION	*/
	int						type;	/* was ist im UNION enthalten?	*/
	union {
		BPath				*bpath;	/* enthält das BPath-Objekt zum Zugriff auf die Datei	*/
		BFile				*bfile;	/* Enthält ein evtl. BFile-Objekt	*/
		BVolumeRoster		*bvolume;	/* Enthält das BVolumeRoster Objekt	*/
		BDirectory			*bdir;	/* Enthält ein BDirectory	*/
	}						pointer;
} OPEN_HANDLER;

static OPEN_HANDLER		open_files[MAX_OPEN_FILES];

#pragma export on
extern "C" __declspec(dllexport) int32 RunServer(void) {

	port_id			pid_o, pid_i;
	unsigned char	*buffer, *b;
	unsigned char	*send_buffer;	// Der Buffer zum senden kann immer unterschiedlich lang sein!
	unsigned char   outb[256];
	char			*c, *c2, *c3;
	int				ergebnis, typ;
	int32			message;
	ssize_t 		p_len, i, send_len;
	BPath			*bp;
	BEntry			*bent;

	int				word1, word2;
	long			long1, long2;
	char			*str1, *str2;
	
#ifdef DEBUG_FILE
	FILE			*fd_debug;
	
	fd_debug = fopen(DEBUG_FILE, "w");
	if (fd_debug != NULL) {
		fprintf(fd_debug, "RFSV-Server gestartet!\n");
		fclose(fd_debug);
	}
#endif

	pid_i = find_port("SYS$RFSV");			// Über diesen Port kommen die zu ankommenden Nachrichten
	pid_o = find_port("WritePsion");		// Über diesen Port gehen die zu versendenden Nachrichten

	for (i = 0; i < MAX_OPEN_FILES; i++) {
		open_files[i].aktiv = false;
		open_files[i].fd = i+1;
	}

	if (pid_o < B_NO_ERROR || pid_i < B_NO_ERROR) exit_thread(pid_o);

	for (i = 0; i < 5; i++) {						// 5 Versuche wg. Rauschen :-)
		read_port(pid_i, &message, NULL, 0);		// die erste Nachricht die eingeht, ist der filedeskriptor!
		if ((message & CMD_MASK) == PSION_START) {
			cid = message & CID_MASK;
			break;									// for-schleife verlassen
		} else printf("SYS$RFSV: Falsche message %lx %lx\n", message & CMD_MASK, PSION_START);
	}
	if (i >= 5) exit_thread(-1);

	send_buffer = NULL;								// Kennzeichnen wir den leeren Buffer!
	
	do {
		p_len = port_buffer_size(pid_i);			// Wie groß ist die Nachricht, die ankommt?
		if (p_len < B_NO_ERROR) exit_thread(p_len);	// Port wurde gelöscht?

		buffer = (unsigned char *)malloc(p_len);	// Speicher für die Nachricht reservieren.
		if (buffer != NULL) {
			read_port(pid_i, &message, buffer, p_len);

#ifdef DEBUG_FILE
	fd_debug = fopen(DEBUG_FILE, "a+");
	if (fd_debug != NULL) {
		fprintf(fd_debug, "\nRFSV-IN :    ");
		for (i = 0; i < p_len; fprintf(fd_debug, "%02x ", buffer[i++])) ;
		fprintf(fd_debug, "\n   ASCII:    ");
		for (i = 0; i < p_len; i++) fprintf(fd_debug, " %c ", buffer[i] >= ' ' ? buffer[i] : '.');
		fprintf(fd_debug, "\n");
		fclose(fd_debug);
	}
#endif

			send_len = -1;
			outb[0] = 1;				// Anzahl der folgenden Frames?
			b = &outb[7];				// Bei Byte 7 beginnt der Inhalt der Daten
			
			typ = BytesToWord(buffer);
			switch ( typ ) {
				case RF_FOPEN:
					word1 = BytesToWord(&buffer[4]);

#ifdef DPRINTF_3
DPRINTF_3("RF_FOPEN >%s< 0x%04x \n", (char *)&buffer[6], word1);
#endif
					send_buffer = (unsigned char *) malloc (2 + MIN_SEND_BUFFER); // fd wird geliefert
					send_len = 2;
					
					ergebnis = open_file((char *)&buffer[6], word1);
					if (ergebnis > 0) {
						WordToBytes(ergebnis, &send_buffer[MIN_SEND_BUFFER]); // Filedeskriptor kopieren.
						ergebnis = 0;
					} else send_len = 0;
					break;
				case RF_FCLOSE:
					word1 = BytesToWord(&buffer[4]);

#ifdef DPRINTF_2
DPRINTF_2("RF_FCLOSE 0x%04x\n", word1);
#endif

					send_buffer = (unsigned char *) malloc (0 + MIN_SEND_BUFFER);
					send_len = 0;
					
					ergebnis = close_file(word1);	// Welchen Kanal schließen?
					break;
				case RF_FREAD:
					word1 = BytesToWord(&buffer[4]);
					word2 = BytesToWord(&buffer[6]);

#ifdef DPRINTF_3
DPRINTF_3("RF_FREAD 0x%04x Anz:%d\n", word1, word2);
#endif

					send_buffer = (unsigned char *)malloc(word2 + MIN_SEND_BUFFER);		// Speicher für den angeforderte Größe + Steuerdaten reservieren!
					ergebnis = read_file(word1, word2, &send_buffer[MIN_SEND_BUFFER]);	// Welchen Kanal lesen und wieviel?
					if (ergebnis > 0) {
						send_len = ergebnis;
						ergebnis = 0;
					} else send_len = 0;
					break;
				case RF_FDIRREAD:
					word1 = BytesToWord(&buffer[4]);
#ifdef DPRINTF_2
DPRINTF_2("RF_FDIRREAD 0x%04x\n", word1);
#endif

					send_buffer = (unsigned char *)malloc(590 + MIN_SEND_BUFFER);			// Speicher für den angeforderte Größe + Steuerdaten reservieren!
					ergebnis = fill_dir_buffer(word1, &send_buffer[MIN_SEND_BUFFER], 590);	// 590 Byte mit Directoryinfos füllen
					if (ergebnis > 0) {
						send_len = ergebnis;
						ergebnis = 0;
					} else send_len = 0;
					break;
				case RF_FDEVICEREAD:		// Info über die verschiedenen Devices erwünscht
					word1 = BytesToWord(&buffer[4]);
#ifdef DPRINTF_2
DPRINTF_2("RF_FDEVICEREAD 0x%04x\n", word1);
#endif
					ergebnis = get_next_device(word1, b);
					if (ergebnis > 0) {
						send_len = ergebnis;
						ergebnis = 0;
					} else {
						send_len = 0;
					}
					break;
				case RF_FWRITE:		// Schreiben der Daten in ein File
					word1 = BytesToWord(&buffer[4]);
#ifdef DPRINTF_2
DPRINTF_2("RF_FWRITE 0x%04x\n", word1);
#endif
					
					send_buffer = (unsigned char *) malloc (0 + MIN_SEND_BUFFER);
					send_len = 0;
					ergebnis = write_file(word1, BytesToWord(&buffer[2])-2, &buffer[6]);

					break;
				case RF_FSEEK:		// Positionieren des Filepointers
					word1 = BytesToWord(&buffer[4]);
					long1 = BytesToLong(&buffer[6]);
					word2 = BytesToWord(&buffer[10]);
#ifdef DPRINTF_4
DPRINTF_4("RF_FSEEK 0x%04x 0x%08x 0x%04x\n", word1, long1, word2);
#endif
					send_buffer = (unsigned char *) malloc (4 + MIN_SEND_BUFFER);
					send_len = 4;
					
					long2 = fseek_file(word1, long1, word2);
					if (long2 >= 0) {
						LongToBytes(long2, &send_buffer[MIN_SEND_BUFFER]);
						ergebnis = 0;
					} else {
						ergebnis = long2;
						send_len = 0;
					}
					break;
				case RF_FFLUSH:		// Schreiben abschliessen und speicher auf Diskette schreiben!
					word1 = BytesToWord(&buffer[4]);
#ifdef DPRINTF_2
DPRINTF_2("RF_FFLUSH 0x%04x\n", word1);
#endif
					
					send_buffer = (unsigned char *) malloc (0 + MIN_SEND_BUFFER);
					send_len = 0;
					ergebnis = 0;	// Wir machen hierbei nichts!

					break;
				case RF_FSETEOF:		// Ende Der Datei auf den übergebenen Wert setzen
					word1 = BytesToWord(&buffer[4]);
					long1 = BytesToLong(&buffer[6]);
#ifdef DPRINTF_3
DPRINTF_3("RF_FSETEOF 0x%04x 0x%08x\n", word1, long1);
#endif
					
					send_buffer = (unsigned char *) malloc (0 + MIN_SEND_BUFFER);
					send_len = 0;
					ergebnis = seteof_file(word1, long1);
					
					break;
				case RF_RENAME:
					str1 = (char *)&buffer[4];
					str2 = (char *)&buffer[5+strlen(str1)];
#ifdef DPRINTF_3
DPRINTF_3("RF_RENAME >%s< >%s<\n", str1, str2);
#endif
					send_buffer = (unsigned char *) malloc (0 + MIN_SEND_BUFFER);
					send_len = 0;

					ergebnis = rename_file(str1, str2);
					break;
				case RF_DELETE:
					str1 = (char *)&buffer[4];
#ifdef DPRINTF_2
DPRINTF_2("RF_DELETE >%s<\n", str1);
#endif
					send_buffer = (unsigned char *) malloc (0 + MIN_SEND_BUFFER);
					send_len = 0;

					ergebnis = delete_file(str1);
					break;
				case RF_FINFO:
					str1 = (char *)&buffer[4];
#ifdef DPRINTF_2
DPRINTF_2("RF_FINFO >%s<\n", str1);
#endif

					send_buffer = (unsigned char *) malloc (16 + MIN_SEND_BUFFER);
					send_len = 16;

					if (str1[strlen(str1)-1] == '/') {
						ergebnis = E_FILE_NAME;	// Kein erlaubter Filename, wenn am Ende "/" !
					} else {
						str2 = find_filename(str1);
						bent = new BEntry(str2);
						ergebnis = get_status(bent, &send_buffer[MIN_SEND_BUFFER]);
						delete bent;
						free(str2);
					}
					if (ergebnis >= 0) {
						ergebnis = 0;
					} else {
						send_len = 0;
					}
					break;
				case RF_SFSTAT:
					word1 = BytesToWord(&buffer[4]);
					word2 = BytesToWord(&buffer[6]);
					str1 = (char *)&buffer[8];
#ifdef DPRINTF_4
DPRINTF_4("RF_SFSTAT %04x %04x >%s<\n", word1, word2, str1);
#endif

					send_buffer = (unsigned char *) malloc (0 + MIN_SEND_BUFFER);
					send_len = 0;

					ergebnis = set_filestat(str1, word1, word2);
					break;
				case RF_MKDIR:
					str1 = (char *)&buffer[4];
#ifdef DPRINTF_2
DPRINTF_2("RF_MKDIR >%s<\n", &buffer[4]);
#endif
					send_buffer = (unsigned char *) malloc (0 + MIN_SEND_BUFFER);
					send_len = 0;

					ergebnis = make_dir(str1);
					break;
				case RF_PARSE:		// Überprüft lediglich eine Pfad auf sinnvoller Struktur
									// und füllt die Längenwerte. macht das aber sehr kompliziert!

					c = (char *)&buffer[4];
					c2 = c+(strlen(c)+1);
					c3 = c2+(strlen(c2)+1);
#ifdef DPRINTF_4
DPRINTF_4("RF_PARSE >%s< >%s< >%s<\n", c, c2, c3);
#endif

					ergebnis = parse_file(c, c2, c3, b);
					send_len = strlen((char *)&b[6]) + 7;
					break;
				case RF_STATUSDEVICE:
#ifdef DPRINTF_2
DPRINTF_2("RF_STATUSDEVICE >%s<\n", &buffer[4]);
#endif

					ergebnis = get_device((char *)&buffer[5], b);	// ohne fuehrenden "/"
					if (ergebnis > 0) {
						send_len = ergebnis;
						ergebnis = 0;
					} else {
						send_len = 0;
					}
					break;
				case RF_PATHTEST:
#ifdef DPRINTF_2
DPRINTF_2("RF_PATHTEST >%s<\n", &buffer[4]);
#endif

					bp = new BPath((char *)&buffer[4], NULL, true);
					if (bp->Path() == NULL) ergebnis = -1;
					else ergebnis = 0;
					send_len = 0;
					delete bp;
					break; 
				case RF_STATUSSYSTEM:
					printf("RF_STATUSSYSTEM\n");
#ifdef DEBUG_FILE
	if ((fd_debug = fopen(DEBUG_FILE, "a+")) != NULL) {
		fprintf(fd_debug, "RF_STATUSSYSTEM\n");
		fclose(fd_debug);
	}
#endif
					// Der PSION möchte Informationen über das Filesystem erhalten
					
					WordToBytes(RFSV_VERSION, b);		// Versionsnummer für Erweiterungen
					WordToBytes(P_FSYSTYPE_HIER, &b[2]);	// 0 = flaches Filesystem; 1 = Hierarchisches FS
					WordToBytes(false, &b[4]);	// Filesystem erlaubt keine Format-Befehle!
					memset(&b[6], 0xff, 26);	// Reserve auf ff setzen!
for (i=0; i < 13; WordToBytes(0, &b[6+((i++) *2)])) ;
					send_len = 32;
					ergebnis = 0;					// Kein Fehler
        	        break;
				case RF_CHANGEDIR:
					word1 = BytesToWord(&buffer[4]);
					str1 = (char *)&buffer[6];
					str2 = (char *)&buffer[7+strlen(str1)];
					if (word1 != P_CD_SUBDIR) str2--;	// Auf das 0-Byte von str1 setzen!
#ifdef DPRINTF_4
DPRINTF_4("RF_CHANGEDIR: 0x%04x >%s< >%s<\n", word1, str1, str2);
#endif
					// Filelänge + 6, weil REM:: noch davor kommt und das 0-Byte danach!
					send_buffer = (unsigned char *) malloc (B_FILE_NAME_LENGTH + 6 + MIN_SEND_BUFFER);
					
					ergebnis = change_dir(word1, str1, str2, (char *)&send_buffer[MIN_SEND_BUFFER]);
					if (ergebnis >= 0) {
						send_len = strlen((char *)&send_buffer[MIN_SEND_BUFFER])+1;
						ergebnis = 0;
					} else send_len = 0;

					break;
				case RF_SFDATE:
					long1 = BytesToLong(&buffer[4]);
					str1 = (char *) &buffer[8];
#ifdef DPRINTF_3
DPRINTF_3("RF_SFDATE: %ld, >%s<\n", long1, str1);
#endif

					ergebnis = set_crdate(str1, long1);
					send_len = 0;
					break;
        	    default:
#ifdef DEBUG_FILE
	if ((fd_debug = fopen(DEBUG_FILE, "a+")) != NULL) {
		fprintf(fd_debug, "UNKNOWN KOMMAND\n");
		fclose(fd_debug);
	}
#endif
   	    			printf("RFSV-Server-in: UNKNOWN: ");
					for (i = 0; i < p_len; printf("%02x ", buffer[i++])) ;
					printf("\n");
					break;
			}

			if (send_buffer == NULL) {		// Zwischenlösung, bis alles auf send_buffer umgestellt ist
				send_buffer = (unsigned char *) malloc (send_len + 7);
				memcpy(send_buffer, outb, send_len + 7);
			}
			if (send_len >= 0) {
	            WordToBytes(ergebnis, &send_buffer[5]);				// Ergebnis 0 = OK
    	        WordToBytes((unsigned int )send_len+2, &send_buffer[3]);	// Länge der Nachricht
        	    WordToBytes(RF_RESPONSE, &send_buffer[1]);

#ifdef DEBUG_FILE
	fd_debug = fopen(DEBUG_FILE, "a+");
	if (fd_debug != NULL) {
		fprintf(fd_debug, "RFSV-OUT: ");
		for (i = 0; i < send_len+7; fprintf(fd_debug, "%02x ", send_buffer[i++])) ;
		fprintf(fd_debug, "\n   ASCII: ");
		for (i = 0; i < send_len+7; i++) fprintf(fd_debug, " %c ", send_buffer[i] >= ' ' ? send_buffer[i] : '.');
		fprintf(fd_debug, "\n");
		fclose(fd_debug);
	}
#endif

            	write_port(pid_o, PSION_CMD + cid, &send_buffer[1], send_len+6);
			}
			
			if (send_buffer != NULL) {	// Wenn Sendepuffer reserviert wurde, dann auch wieder freigeben!
				free(send_buffer);
				send_buffer = NULL;
			}
			free(buffer);						// Empfangspuffer wird immer reserviert!
		}
	} while (true);
}

#pragma export off

/* Öffnet den angeforderten Ordner / die angeforderte Datei und liefert ein handle zurück	*/

int open_file(const char *name, int status) {
	BPath	*bp;
	int i, erg;
	uint32	mode;
	char	*str;
	
	erg = E_FILE_ALLOC;
	for (i = 0; i < MAX_OPEN_FILES; i++) {
		if (!open_files[i].aktiv) {		// Freien Eintrag gefunden!
			switch (status & P_FMT_MASK) {
				case P_FDEVICE:
					open_files[i].aktiv = true;
					open_files[i].type = OP_BVOL;
					open_files[i].pointer.bvolume = new BVolumeRoster();
					erg = open_files[i].fd;
					break;
				case P_FDIR:
					bp = new BPath(name);
					bp->GetParent(bp);
					open_files[i].aktiv = true;
					open_files[i].type = OP_BDIR;
					open_files[i].pointer.bdir = new BDirectory(bp->Path());
					erg = open_files[i].fd;
					break;
				case P_FTEXT:			// Die Unterschiede sind mir derzeit nicht bekannt!
				case P_FSTREAM:
				case P_FSTREAM_TEXT:
					switch(status & P_FSER_MASK) {
						case P_FOPEN:		mode = B_READ_WRITE; break;
						case P_FCREATE:		mode = B_READ_WRITE | B_CREATE_FILE; break;
						case P_FREPLACE:	mode = B_READ_WRITE | B_CREATE_FILE | B_ERASE_FILE; break;
						case P_FAPPEND:		mode = B_READ_WRITE | B_OPEN_AT_END; break;
				//		case P_FUNIQUE 0x0004 /* Unique file open */ ????
						default: mode = B_READ_ONLY; break;
					}
					open_files[i].aktiv = true;
					open_files[i].type = OP_BFILE;
					
					str = find_filename(name);
					open_files[i].pointer.bfile = new BFile(str, mode);
					free(str);
					
					switch (open_files[i].pointer.bfile->InitCheck()) {
						case B_NO_ERROR:	erg = open_files[i].fd; break;
						default: erg = E_FILE_ACCESS; break;
					}
					break;
				default:
					printf("Unbekanntes Öffnen %04x\n", (status & P_FMT_MASK));
					break;
			}
			break;		// for-Schleife verlassen!
		}
	}
	return(erg);
}

/* Schließt den angeforderten Filediskriptor */

int close_file(int fd) {
	int erg = E_FILE_ALLOC;
	if (open_files[fd-1].aktiv) {
		open_files[fd-1].aktiv = false;
		switch(open_files[fd-1].type) {
			case OP_BVOL:	delete open_files[fd-1].pointer.bvolume;
							erg = 0;
							break;
			case OP_BDIR:	delete open_files[fd-1].pointer.bdir;
							erg = 0;
							break;
			case OP_BFILE:	delete open_files[fd-1].pointer.bfile;
							erg = 0;
							break;
		}
	}
	return(erg);
}

/* Holt die Information zu einem bestimmten Device */

int get_device(char *device, unsigned char *buffer) {
	BVolumeRoster	brost;
	BVolume			bvol;
	char			device_name[128];
	char 			*c;
	status_t		erg;
	
	if ((c = strchr(device, '/')) != NULL) *c = '\0';
	
	while ((erg = brost.GetNextVolume(&bvol)) == B_NO_ERROR) {
		bvol.GetName(device_name);
		if (strcmp(device_name, device) == 0) {
			get_dev_stat(&bvol, buffer, NULL);		// Statusbuffer füllen aber Pfad uninteressant
			return (64); 							// Struktur ist 64 Byte groß
		}
	}
	if (strcmp(device, "boot") == 0) {				// Der PSION will was vom Bootdevice wissen
		brost.GetBootVolume(&bvol);
		get_dev_stat(&bvol, buffer, NULL);
		return(64);
	}
	
	return (E_FILE_EOF);
}

/* Holt die nächste Information aus BVolume	*/

int get_next_device(int fd, unsigned char * buffer) {
	BVolume			bvol;
	status_t		erg;
	char			device[128];
	
	memset((char *)buffer, 0xff, 64);
	if (!open_files[fd-1].aktiv) return(-1);
	if (open_files[fd-1].type != OP_BVOL) return(-1);
	
	while ((erg = open_files[fd-1].pointer.bvolume->GetNextVolume(&bvol)) == B_NO_ERROR) {
		bvol.GetName(device);
		if (strlen(device) != 0) {
			get_dev_stat(&bvol, buffer, (char *)&buffer[64]);			// Statusbuffer füllen
			return (64 + 1 + strlen((char *)&buffer[64])); // 0-Byte nicht vergessen zum mitzaehlen!
		}
	}
	return (E_FILE_EOF);
}

/* Füllt den Status des Devices in den Buffer	*/

void get_dev_stat(BVolume *bvol, unsigned char *buffer, char *dir) {
	BDirectory		bdir;
	BEntry			bent;
	BPath			bpath;
	int stat;
	char			volume[128];
	
	WordToBytes(RFSV_VERSION, buffer);	
	stat = P_FMEDIA_HARDDISK;
	if (bvol->IsReadOnly()) stat = P_FMEDIA_WRITEPROTECTED;
	else if (bvol->IsRemovable()) stat = P_FMEDIA_FLOPPY;
	else if (bvol->IsPersistent()) stat = P_FMEDIA_HARDDISK;
	WordToBytes(stat, &buffer[2]);	// Mediatyp
	WordToBytes(bvol->IsRemovable(), &buffer[4]);	// Kann das Medium entfernt werden?
	LongToBytes(bvol->Capacity(), &buffer[6]);	// Größe in Bytes
	LongToBytes(bvol->FreeBytes(), &buffer[10]);	// Freier Speicherplatz

	bvol->GetName(volume);
	buffer[14] = '\0';
	strncat((char *)&buffer[14], volume, P_VOLUMENAME-1);	// Devicenamen verwenden

	if (dir != NULL) {										// Pfad feststellen!
		if (bvol->GetRootDirectory(&bdir) >= B_NO_ERROR) {
			if (bdir.GetEntry(&bent) >= B_NO_ERROR) {
				if (bent.GetPath(&bpath) >= B_NO_ERROR) {
					strcpy(dir, bpath.Path());				// Laufwerksnamen kopieren
				}
			}
		}
	}
}

/* Füllt einen Buffer mit den Directoryeinträgen!		*/
/* len gibt die Größe des Buffers an.					*/
/* Liefert die Länge des gefüllten Buffers				*/

int fill_dir_buffer(int fd, unsigned char * buffer, size_t len) {
	BEntry				bent;
	BPath				bpath;
	unsigned char *		b;
	
	if (!open_files[fd-1].aktiv) return(-1);
	if (open_files[fd-1].type != OP_BDIR) return(-1);
	
	b = &buffer[2];
	while (open_files[fd-1].pointer.bdir->GetNextEntry(&bent) >= B_NO_ERROR) {
		bent.GetPath(&bpath);
		get_status(&bent, b);
		bent.GetName((char *)&b[16]);

		b = &b[17] + strlen((char *)&b[16]);	// 0-Byte muß stehen bleiben!
		if (b-buffer >= len) break;
	}
	if (b == &buffer[2]) return(E_FILE_EOF);	// Kein Eintrag mehr!

	WordToBytes((b-buffer)-2, buffer);
	return (b-buffer);
}

/* holt den Status der Datei/Directory */

int get_status(BEntry *bent, unsigned char *buffer) {
	struct stat			st;
	int					status;
	BEntry				parent;
	char name[B_FILE_NAME_LENGTH]; 
	
	WordToBytes(RFSV_VERSION, buffer);

	bent->GetParent(&parent);
	if (!parent.Exists()) return(E_FILE_DIR);	// Directory existiert nicht!
	if (!bent->Exists()) return(E_FILE_NXIST);	// File gibt es nicht
	bent->GetStat(&st);
	
	parent.GetName(name);	
	if (strcmp(name, ".") == 0) return(E_FILE_NAME);
	
	status=P_FAREAD|P_FAWRITE;
	if (S_ISDIR(st.st_mode)) status |= P_FADIR;
	else status |= P_FASTREAM;
	
	WordToBytes(status, &buffer[2]);
		
	if (status & P_FADIR) LongToBytes(0, &buffer[4]);
	else LongToBytes(st.st_size, &buffer[4]);		// Größe feststellen
	LongToBytes(st.st_mtime, &buffer[8]);		// Modificationtime
		
	memset(&buffer[12], 0xff, 4);	// Spare
	return(0);
}

/* Überprüft ob der Dateiname ein gültiger Dateiname ist	*/
/* REM::/BeDisk/ordner1/ordner2/datei*.ex*"                 */
/* --5-!---8---!------16-------!--6--!-4-!					*/
/* Filesystem und Device können weggelassen werden, dann ist das Standardlaufwerk einzusetzen */

int parse_file(char *file1, char *file2, char *file3, unsigned char *werte) {
	char	*f[3];
	char	*ergebnis;
	BEntry		bent;
	int i, len;
	
	f[0] = file1;
	f[1] = file2;
	f[2] = file3;
	memset(werte, 0x00, 6);
	ergebnis = (char *)&werte[6];
	for (i = 0; i < 3; i++) if ((len = copy_filesystem(f[i], ergebnis)) > 0) break;	// Filesystem suchen
	if (i >= 3) {	// Kein Filesystem übegeben, also Standard eintragen!
		strcpy(ergebnis, "REM::");
		ergebnis += 5;
		werte[0] = 5;
	} else {
		ergebnis += len;
		werte[0] = len;
		f[i] += len;
		for (i++ ; i < 3; i++) {						// aus allen anderen Strings das Filesystem entfernen!
			len = copy_filesystem(f[i], NULL);
			if (len > 0) f[i] += len;
		}
	}
	for (i = 0; i < 3; i++) if ((len = copy_device(f[i], ergebnis)) > 0) break;	// Device suchen
	if (i >= 3) {	// Kein Device übegeben, also Standard eintragen!
		strcpy(ergebnis, "/boot");
		ergebnis += 5;
		werte[1] = 5;
	} else {
		ergebnis += len;
		werte[1] = len;
		f[i] += len;
		for (i++ ; i < 3; i++) {						// aus allen anderen Strings das Filesystem entfernen!
			len = copy_device(f[i], NULL);
			if (len > 0) f[i] += len;
		}
	}

	for (i = 0; i < 3; i++) if ((len = copy_path(f[i], ergebnis)) > 0) break;	// Pfad suchen
	if (i >= 3) {	// Kein Pfad übegeben, also Standard eintragen, d.h. kein Pfad
		strcpy(ergebnis, "/");
		ergebnis++;
		werte[2] = 1;
	} else {
		ergebnis += len;
		werte[2] = len;
		f[i] += len;
		for (i++ ; i < 3; i++) {						// aus allen anderen Strings das Filesystem entfernen!
			len = copy_path(f[i], NULL);
			if (len > 0) f[i] += len;
		}
	}

	for (i = 0; i < 3; i++) if ((len = copy_filename(f[i], ergebnis)) > 0) break;	// Filename suchen
	if (i >= 3) {	// Kein Filename übegeben, also nix eintragen und wert3 auf 0 setzen!!
		werte[3] = 0;
		werte[5] = 0; // dann auch kein Wildcard enthalten!
	} else {
		if (strchr(ergebnis, '*') != NULL)  werte[5] |= (P_PWILD_ANY | P_PWILD_NAME); 
		ergebnis += len;
		werte[3] = len;
		f[i] += len;
		for ( ; i < 3; i++) {						// aus allen anderen Strings den Filenamen entfernen!
			len = copy_filename(f[i], NULL);
			if (len > 0) f[i] += len;
		}
	}
	
	for (i = 0; i < 3; i++) {
		if (*f[i] == '\0') continue;
		strcpy(ergebnis, f[i]);
		if (strcmp(f[i], ".*") != 0) break;
	}
	werte[4] = strlen(ergebnis);
	if (strchr(ergebnis, '*') != NULL)  werte[5] |= (P_PWILD_ANY | P_PWILD_EXT);

	return 0;
}

// Überprüft, ob im String "in" ein Filesystem enthalten ist. 
// Wenn ja, dann wird dieser String nach filesys kopiert (wenn filesys != NULL)
// Der Rückgabewert ist die Anzahl der Zeichen die das Filesystem hat.
// Wenn kein FS gefunden, dann wird -1 zurückgegeben!

int copy_filesystem(const char *in, char *filesys) {
	if (strncmp(in, "REM::", 5) == 0) {
		if (filesys != NULL) strcpy(filesys, "REM::");
		return 5;
	} else return -1;
}

// Überprüft, ob im String "in" ein Device enthalten ist. 
// Wenn ja, dann wird dieser String nach device kopiert (wenn device != NULL)
// Der Rückgabewert ist die Anzahl der Zeichen die das Filesystem hat.
// Wenn kein FS gefunden, dann wird -1 zurückgegeben!

int copy_device(const char *in, char *device) {
	BVolumeRoster	brost;
	BVolume			bvol;
	size_t			len;
	char			device_name[128];

	if (*in++ == '/') {	// Ein Device muß mit "/" anfangen!
		
		while (brost.GetNextVolume(&bvol) == B_NO_ERROR) {
			bvol.GetName(device_name);
			len = strlen(device_name);
			if (len <= 0) continue;
			
			if (strncmp(device_name, in, len) == 0) {
				if (device != NULL) {
					*device = '/';
					strcpy(&device[1], device_name);
				}
				return len + 1;		// Ersten "/" mitzaehlen!
			}
		}
		if (strncmp(in, "boot", 4) == 0) {
			if (device != NULL) strcpy(device, "/boot");
			return 5;
		}
	} 
	return -1;
}


// Überprüft, ob im String "in" ein path enthalten ist. 
// Wenn ja, dann wird dieser String nach path kopiert (wenn path != NULL)
// Der Rückgabewert ist die Anzahl der Zeichen die der Path hat.
// Wenn kein path gefunden, dann wird -1 zurückgegeben!

int copy_path(const char *in, char *path) {
	char *c;
	size_t len;
	
	c = strrchr(in, '/');
	if (c != NULL) {
		len = (c-in)+1;
		if (path != NULL) {
			*path = '\0';
			strncat(path, in, len);
		}
		return len;
	}
	return -1;
}

// Überprüft, ob im String "in" ein filename enthalten ist. 
// Wenn ja, dann wird dieser String nach filename kopiert (wenn filename != NULL)
// Der Rückgabewert ist die Anzahl der Zeichen die der filename hat.

int copy_filename(const char *in, char *filename) {
	char *c;
	size_t len;
	
	c = strchr(in, '.');
	if (c != NULL) len = (c-in);
	else len = strlen(in);
	
	if (filename != NULL) {
		*filename = '\0';
		strncat(filename, in, len);
	}
	return len;
}

// erstellt ein Verzeichnis

int make_dir(const char *dirname) {
	BDirectory bdir, neues_bdir;
	status_t	erg;
	char		*c;
	
	c = strrchr(dirname, '/');					// Am Ende des Pfades darf kein "/" stehen, sonst kann der Ordner nicht angelegt werden!
	if (c != NULL && c[1] == '\0') *c = '\0';	// Daher entfernen wir einen evtl. "/" vom Ende. Ich wunder mich, daß das trotz
												// const compiliert werden kann!
	erg = bdir.CreateDirectory(dirname, &neues_bdir);
	if (erg < B_NO_ERROR) return (E_FILE_ACCESS);
	else return 0;
}

// verschiebt eine Datei/Ordner in den Mülleimer

int delete_file(const char *filename) {
	BEntry		bent;
	BPath		path;
	BDirectory	dir;
	status_t	erg;
	char		*file;
	
	file = find_filename(filename);
	bent.SetTo(file);
	free(file);
	
	if (bent.Exists()) {
  		find_directory(B_TRASH_DIRECTORY, &path, true);	// Wo ist der Mülleimer
		dir.SetTo(path.Path());
		erg = bent.MoveTo(&dir, NULL, true);
		switch (erg) {
			case B_NO_ERROR: return 0;
			case B_NO_INIT:printf("Ergebnis = B_NO_INIT\n");break;
			case B_ENTRY_NOT_FOUND:printf("Ergebnis = B_ENTRY_NOT_FOUND\n");break;
			case B_FILE_EXISTS:printf("Ergebnis = B_FILE_EXISTS\n");break;
			case B_BUSY:printf("Ergebnis = B_BUSY\n");break;
		}
	}
	return (E_FILE_NXIST);
}

// verschiebt eine Datei/Ordner

int rename_file(const char *filename_alt, const char *filename_neu) {
	BEntry		bent_alt, bent_neu;
	BDirectory	dir;
	char		leaf[128];
	status_t	erg;
	char		*fa, *fn;
	
	fa = find_filename(filename_alt);
	fn = find_filename(filename_neu);

	bent_alt.SetTo(fa);
	bent_neu.SetTo(fn);
	if (bent_alt.Exists()) printf("ALT Existiert!\n");
	if (bent_neu.Exists()) printf("NEU existiert!\n");
	if (bent_alt.Exists() && !bent_neu.Exists()) {
		bent_neu.GetParent(&dir);
		bent_neu.GetName(leaf);
		erg = bent_alt.MoveTo(&dir, leaf);
		switch (erg) {
			case B_NO_ERROR: return 0;
			case B_NO_INIT:printf("Ergebnis = B_NO_INIT\n");break;
			case B_ENTRY_NOT_FOUND:printf("Ergebnis = B_ENTRY_NOT_FOUND\n");break;
			case B_FILE_EXISTS:printf("Ergebnis = B_FILE_EXISTS\n");break;
			case B_BUSY:printf("Ergebnis = B_BUSY\n");break;
			default: printf("Ergebnis = %lx\n", erg);break;
		}
	}
	return (E_FILE_NXIST);
}

// gibt als Information zurück, wie ein bestimmter Pfad als Ergebnis aussehen würde.

int change_dir(int art, const char *start, const char *zusatz, char *ergebnis) {
	char	*filenam, *c;

	filenam = strrchr(start, '/');
	if (filenam != NULL ) {
		*filenam++ = '\0';		// start auf Pfadnamen beschränken
	}

	strcpy(ergebnis, "REM::");
	
	switch(art) {
		case P_CD_ROOT:
			c = strchr((char *)&start[1], '/');
			
			if (c != NULL) *c = '\0';
			strcpy(&ergebnis[5], start);
			break;
		case P_CD_PARENT:

			c = strrchr(start, '/');
			if (c != NULL && c != start) {
				*c = '\0';
				strcpy(&ergebnis[5], start);
			} else return E_FILE_NXIST;			// geht nicht höher!
			
			break;
		case P_CD_SUBDIR:
			strcpy(&ergebnis[5], start);
			strcat(&ergebnis[5], "/");
			strcat(&ergebnis[5], zusatz);
			break;
	}
	strcat(&ergebnis[5], "/");
	strcat(&ergebnis[5], filenam);
	return 0;
}

// Macht den fseek für eine Datei

int32 fseek_file(int fd, long anz, int mode) {
	int m;
	
	switch(mode) {
		case P_FABS:	m = SEEK_SET; break;
		case P_FEND:	m = SEEK_END; break;
		case P_FCUR:	m = SEEK_CUR; break;
		case P_FRSENSE: /* Sense the record position */
		case P_FRSET: /* Set the record position */
		case P_FREWIND: /* Rewind a text file */
			printf("Dont Know, what to do with mode: %d\n", mode);
			return -1;
	}
	if (open_files[fd-1].aktiv) {
		if (open_files[fd-1].type == OP_BFILE) {
			open_files[fd-1].pointer.bfile->Seek(anz, m);
			return (open_files[fd-1].pointer.bfile->Position());
		}
	}
	return -1;
}

// Schreibt Daten in den übergebenen FD

int write_file(int fd, int anz, const unsigned char *buffer) {

	ssize_t erg = E_FILE_WRITE;
	
	if (open_files[fd-1].aktiv) {
		if (open_files[fd-1].type == OP_BFILE) {
			erg = open_files[fd-1].pointer.bfile->Write(buffer, anz);
		}
	}
	if (erg < 0) {
		switch (erg) {
			case B_DEVICE_FULL: return (E_FILE_FULL); break;
			case B_NOT_ALLOWED: return (E_FILE_RDONLY); break;
			default:			return erg;
		}
	} else return(0);
}

// Liest Daten aus dem übergebenen FD

int read_file(int fd, int anz, unsigned char *buffer) {

	ssize_t erg = E_FILE_WRITE;
	if (buffer == NULL) return -1;
	
	if (open_files[fd-1].aktiv) {
		if (open_files[fd-1].type == OP_BFILE) {
			erg = open_files[fd-1].pointer.bfile->Read(buffer, anz);
		}
	}
	if (erg == 0) return (E_FILE_EOF);
	else return erg;
}

// Setzt das Ende der Datei

int seteof_file(int fd, long laenge) {

	ssize_t erg = E_FILE_WRITE;

	if (open_files[fd-1].aktiv) {
		if (open_files[fd-1].type == OP_BFILE) {
			erg = open_files[fd-1].pointer.bfile->SetSize(laenge);
		}
	}
	return 0;
}

// Diese Funktion muß die entsprechenden Flags setzen!
// Es werden alle Bits bearbeitet, die in MASK gesetzt wurden!
// Problem ergibt sich wieder mit groß/kleinschreibung des Filenamens!

int set_filestat(const char *filename, int bit, int mask) {
	BNode node;
	mode_t permission;
	char	*file;
	int erg = E_FILE_NXIST;
	
	file = find_filename(filename);
	if (node.SetTo(file) >= B_NO_ERROR) {
		permission = S_IRUSR;
		if ((mask & P_FAWRITE) && (bit & P_FAWRITE)) permission |= S_IWUSR;
		if ((mask & P_FAMOD) && (bit & P_FAMOD)) node.SetModificationTime(real_time_clock());
		erg = 0;
	}
	free(file);
	return erg;	
}

// Diese Funktion setzt das Erstellungsdatum!
// Problem ergibt sich wieder mit groß/kleinschreibung des Filenamens!

int set_crdate(const char *filename, long secs) {
	BNode node;
	char	*file;
	int 	erg = E_FILE_NXIST;
	
	file = find_filename(filename);
	if (node.SetTo(filename) >= B_NO_ERROR) {
		node.SetCreationTime(secs);
		erg = 0;
	}
	free(file);
	return erg;
}

// Da das Filesystem des PSION nicht casesensitive arbeitet, muss zuerst der richtige Dateinamen
// gesucht werden. Dazu wird in folgenden Schritten vorgegangen:
// 1. Name vorhanden, so wie angegeben?		ja-> Name zurückgeben
// 2. uebergebener Name mit strcasecmp vergleichen.
// wird der Name nicht gefunden, so wird ein kleingeschriebener Name zurückgeben!

char *find_filename(const char *filename) {
	BDirectory	bdir;
	BEntry		bent;
	BPath		path;
	status_t	erg;
	char		*file_erg;
	char		file_pc[B_FILE_NAME_LENGTH], file_psion[B_FILE_NAME_LENGTH], *c;
	
	file_erg = (char *) malloc(B_FILE_NAME_LENGTH);
	strcpy(file_erg, filename);		// Filenamen kopieren um damit arbeiten zu können.

	bent.SetTo(file_erg);
	bent.GetName(file_psion);
	bent.GetParent(&bdir);
	while ((erg = bdir.GetNextEntry(&bent)) >= B_NO_ERROR) {
		bent.GetName(file_pc);
		if (strcasecmp(file_pc, file_psion) == 0) {
//			printf("Datei gefunden! Angefordert: %s gefunden %s\n", file_psion, file_pc);
			bent.GetPath(&path);
			strcpy(file_erg, path.Path());
//			printf("Path:org:%s\n     erg:%s\n", filename, path.Path());
			return(file_erg);
		}
	}
//	printf("\n");
	c = strrchr(file_erg, '/');
	if (c == NULL) c = file_erg;	// wenn kein "/" im Dateiname, dann den ganzen Dateinamen bearbeiten!
	for (; *c != '\0'; c++) if (*c < 0x80 && *c >= 0x40)  *c |= 0x20;
	return (file_erg);
}

void WordToBytes(unsigned int i, unsigned char *p) {
    *p++ = i;
    *p = (i >> 8);
}

void LongToBytes(unsigned long l, unsigned char *p) {
    *p++ = l;
    *p++ = (l >> 8);
    *p++ = (l >> 16);
    *p = (l >> 24);
}

unsigned int BytesToWord(const unsigned char *c) {
	return ((unsigned int )*c | ((unsigned int )c[1] << 8)); 
}

unsigned long BytesToLong(const unsigned char *c) {
	return ((unsigned int )*c | ((unsigned int )c[1] << 8) | ((unsigned int )c[2] << 16) | ((unsigned int )c[3] << 24)); 
}

